import 'dart:io';

import 'package:dio/dio.dart';
import 'package:soliq_id/soliq_id.dart';
import 'package:soliq_id/src/data/model/response/response_model.dart';

enum NetworkStatus {
  initial,
  loading,
  success,
  failure,
  refresh,
}

class NetworkService {
  Future<NetworkResponseModel> getMethod({
    required String url,
    bool hasHeader = true,
    Map<String, dynamic>? queryParameters,
  }) async {
    if (hasHeader) {
      final header = _getHeader();
      dio?.options.headers = header;
    } else {
      dio?.options.headers = {};
    }
    try {
      final response = await dio?.get(url, queryParameters: queryParameters);
      return NetworkResponseModel.success(response: response);
    } on DioException catch (error) {
      if (error.type == DioExceptionType.connectionTimeout) {
        return NetworkResponseModel.error(
          errorMessage: error.message,
        );
      } else {
        return NetworkResponseModel.error(
          errorMessage: error.message,
        );
      }
    } catch (error) {
      return NetworkResponseModel.error(
        errorMessage: error.toString(),
      );
    }
  }

  Future<NetworkResponseModel> postMethod({
    required String url,
    required dynamic body,
    bool hasHeader = true,
  }) async {
    if (hasHeader) {
      final header = _getHeader();
      dio?.options.headers = header;
    } else {
      dio?.options.headers = {};
    }
    try {
      final response = await dio?.post(
        url,
        data: body,
      );

      return NetworkResponseModel.success(response: response);
    } on DioException catch (error) {
      if (error.type == DioExceptionType.connectionTimeout) {
        return NetworkResponseModel.error(
          errorMessage: error.message,
        );
      } else {
        return NetworkResponseModel.error(
          errorMessage: error.message,
        );
      }
    } catch (error) {
      return NetworkResponseModel.error(
        errorMessage: error.toString(),
      );
    }
  }

  Future<NetworkResponseModel> putMethod({
    required String url,
    required dynamic body,
    bool hasHeader = true,
  }) async {
    if (hasHeader) {
      final header = _getHeader();
      dio?.options.headers = header;
    } else {
      dio?.options.headers = {};
    }
    try {
      final response = await dio?.put(
        url,
        data: body,
      );

      return NetworkResponseModel.success(response: response);
    } on DioException catch (error) {
      if (error.type == DioExceptionType.connectionTimeout) {
        return NetworkResponseModel.error(
          errorMessage: error.message,
        );
      } else {
        return NetworkResponseModel.error(
          errorMessage: error.message,
        );
      }
    } catch (error) {
      return NetworkResponseModel.error(
        errorMessage: error.toString(),
      );
    }
  }

  Map<String, String> _getHeader() {
    final Map<String, String> headers = {
      'Content-Type': 'application/json',
      'charset': 'utf-8',
      HttpHeaders.authorizationHeader: "Bearer $guid",
    };

    return headers;
  }

  Future<NetworkResponseModel> filePostMethod({
    required String url,
    required dynamic body,
    bool hasHeader = true,
  }) async {
    try {
      dio?.options.headers = fileHeader(hasHeader: hasHeader);

      final response = await dio?.post(
        url,
        data: body,
      );

      return NetworkResponseModel.success(response: response);
    } on DioException catch (error) {
      if (error.type == DioExceptionType.connectionTimeout) {
        return NetworkResponseModel.error(
          errorMessage: error.message,
        );
      } else {
        return NetworkResponseModel.error(
          errorMessage: error.message,
        );
      }
    } catch (error) {
      return NetworkResponseModel.error(
        errorMessage: error.toString(),
      );
    }
  }

  Map<String, String> fileHeader({required bool hasHeader}) {
    final Map<String, String> headers = {
      'Content-Type': 'application/json',
      'charset': 'utf-8',
      HttpHeaders.authorizationHeader: "Bearer $guid",
    };

    return headers;
  }
}
